@@ -1,5 +1,6 @@
 from django.core.cache.backends.base import BaseCache, InvalidCacheBackendError
 from django.core.exceptions import ImproperlyConfigured
+from django.utils import importlib
 from django.utils.encoding import smart_unicode, smart_str
 from django.utils.datastructures import SortedDict
 
@@ -13,8 +14,8 @@ try:
 except ImportError:
     raise InvalidCacheBackendError(
         "Redis cache backend requires the 'redis-py' library")
-
-
+from redis.connection import UnixDomainSocketConnection, Connection
+from redis.connection import DefaultParser
 
 class CacheKey(object):
     """
@@ -29,34 +30,123 @@ class CacheKey(object):
     def __str__(self):
         return self.__unicode__()
 
+    def __repr__(self):
+        return self.__unicode__()
+
     def __unicode__(self):
         return smart_str(self._key)
 
 
+class CacheConnectionPool(object):
+    _connection_pool = None
+
+    def get_connection_pool(self, host='127.0.0.1', port=6379, db=1,
+        password=None, parser_class=None,
+        unix_socket_path=None):
+        if self._connection_pool is None:
+            connection_class = (
+                unix_socket_path and UnixDomainSocketConnection or Connection
+            )
+            kwargs = {
+                'db': db,
+                'password': password,
+                'connection_class': connection_class,
+                'parser_class': parser_class,
+            }
+            if unix_socket_path is None:
+                kwargs.update({
+                    'host': host,
+                    'port': port,
+                })
+            else:
+                kwargs['path'] = unix_socket_path
+            self._connection_pool = redis.ConnectionPool(**kwargs)
+        return self._connection_pool
+pool = CacheConnectionPool()
+
+
 class CacheClass(BaseCache):
     def __init__(self, server, params):
         """
         Connect to Redis, and set up cache backend.
         """
+        self._init(server, params)
+
+    def _init(self, server, params):
         super(CacheClass, self).__init__(params)
-        options = params.get('OPTIONS', {})
-        password = params.get('password', options.get('PASSWORD', None))
-        db = params.get('db', options.get('DB', 1))
-        try:
-            db = int(db)
-        except (ValueError, TypeError):
-            raise ImproperlyConfigured("db value must be an integer")
-        if ':' in server:
-            host, port = server.split(':')
+        self._server = server
+        self._params = params
+
+        unix_socket_path = None
+        if ':' in self.server:
+            host, port = self.server.split(':')
             try:
                 port = int(port)
             except (ValueError, TypeError):
                 raise ImproperlyConfigured("port value must be an integer")
         else:
-            host = server or 'localhost'
-            port = 6379
-        self._cache = redis.Redis(host=host, port=port, db=db, password=password)
+            host, port = None, None
+            unix_socket_path = self.server
 
+        kwargs = {
+            'db': self.db,
+            'password': self.password,
+            'host': host,
+            'port': port,
+            'unix_socket_path': unix_socket_path,
+        }
+        connection_pool = pool.get_connection_pool(
+            parser_class=self.parser_class,
+            **kwargs
+        )
+        self._client = redis.Redis(
+            connection_pool=connection_pool,
+            **kwargs
+        )
+
+    @property
+    def server(self):
+        return self._server or "127.0.0.1:6379"
+
+    @property
+    def params(self):
+        return self._params or {}
+
+    @property
+    def options(self):
+        return self.params.get('OPTIONS', {})
+
+    @property
+    def db(self):
+        _db = self.params.get('db', self.options.get('DB', 1))
+        try:
+            _db = int(_db)
+        except (ValueError, TypeError):
+            raise ImproperlyConfigured("db value must be an integer")
+        return _db
+
+    @property
+    def password(self):
+        return self.params.get('password', self.options.get('PASSWORD', None))
+
+    @property
+    def parser_class(self):
+        cls = self.options.get('PARSER_CLASS', None)
+        if cls is None:
+            return DefaultParser
+        mod_path, cls_name = cls.rsplit('.', 1)
+        try:
+            mod = importlib.import_module(mod_path)
+            parser_class = getattr(mod, cls_name)
+        except (AttributeError, ImportError):
+            raise ImproperlyConfigured("Could not find parser class '%s'" % parser_class)
+        return parser_class
+
+    def __getstate__(self):
+        return {'params': self._params, 'server': self._server}
+
+    def __setstate__(self, state):
+        self._init(**state)
 
     def make_key(self, key, version=None):
         """
@@ -74,7 +164,7 @@ class CacheClass(BaseCache):
         Returns ``True`` if the object was added, ``False`` if not.
         """
         key = self.make_key(key, version=version)
-        if self._cache.exists(key):
+        if self._client.exists(key):
             return False
         return self.set(key, value, timeout)
 
@@ -85,23 +175,31 @@ class CacheClass(BaseCache):
         Returns unpickled value if key is found, the default if not.
         """
         key = self.make_key(key, version=version)
-        value = self._cache.get(key)
+        value = self._client.get(key)
         if value is None:
             return default
-        return self.unpickle(value)
+        try:
+            result = int(value)
+        except (ValueError, TypeError):
+            result = self.unpickle(value)
+        return result
+
 
     def set(self, key, value, timeout=None, version=None, client=None):
         """
         Persist a value to the cache, and set an optional expiration time.
         """
         if not client:
-            client = self._cache
+            client = self._client
         key = self.make_key(key, version=version)
-        # get timout
         if not timeout:
             timeout = self.default_timeout
-        # store the pickled value
-        result = self._cache.setex(key, pickle.dumps(value), int(timeout))
+        try:
+            value = int(value)
+        except (ValueError, TypeError):
+            result = client.setex(key, pickle.dumps(value), int(timeout))
+        else:
+            result = client.setex(key, value, int(timeout))
         # result is a boolean
         return result
 
@@ -109,7 +207,7 @@ class CacheClass(BaseCache):
         """
         Remove a key from the cache.
         """
-        self._cache.delete(self.make_key(key, version=version))
+        self._client.delete(self.make_key(key, version=version))
 
     def delete_many(self, keys, version=None):
         """
@@ -117,14 +215,14 @@ class CacheClass(BaseCache):
         """
         if keys:
             keys = map(lambda key: self.make_key(key, version=version), keys)
-            self._cache.delete(*keys)
+            self._client.delete(*keys)
 
     def clear(self):
         """
         Flush all cache keys.
         """
         # TODO : potential data loss here, should we only delete keys based on the correct version ?
-        self._cache.flushdb()
+        self._client.flushdb()
 
     def unpickle(self, value):
         """
@@ -142,11 +240,14 @@ class CacheClass(BaseCache):
         recovered_data = SortedDict()
         new_keys = map(lambda key: self.make_key(key, version=version), keys)
         map_keys = dict(zip(new_keys, keys))
-        results = self._cache.mget(new_keys)
+        results = self._client.mget(new_keys)
         for key, value in zip(new_keys, results):
             if value is None:
                 continue
-            value = self.unpickle(value)
+            try:
+                value = int(value)
+            except (ValueError, TypeError):
+                value = self.unpickle(value)
             if isinstance(value, basestring):
                 value = smart_unicode(value)
             recovered_data[map_keys[key]] = value
@@ -160,11 +261,27 @@ class CacheClass(BaseCache):
         If timeout is given, that timeout will be used for the key; otherwise
         the default cache timeout will be used.
         """
-        pipeline = self._cache.pipeline()
+        pipeline = self._client.pipeline()
         for key, value in data.iteritems():
             self.set(key, value, timeout, version=version, client=pipeline)
         pipeline.execute()
 
+    def incr(self, key, delta=1, version=None):
+        """
+        Add delta to value in the cache. If the key does not exist, raise a
+        ValueError exception.
+        """
+        key = self.make_key(key, version=version)
+        exists = self._client.exists(key)
+        if not exists:
+            raise ValueError("Key '%s' not found" % key)
+        try:
+            value = self._client.incr(key, delta)
+        except redis.ResponseError:
+            value = self.get(key) + 1
+            self.set(key, value)
+        return value
+
 
 class RedisCache(CacheClass):
     """
@@ -188,7 +305,7 @@ class RedisCache(CacheClass):
             version = self.version
         old_key = self.make_key(key, version)
         value = self.get(old_key, version=version)
-        ttl = self._cache.ttl(old_key)
+        ttl = self._client.ttl(old_key)
         if value is None:
             raise ValueError("Key '%s' not found" % key)
         new_key = self.make_key(key, version=version+delta)
